;;
;        Filename  : RShand.asm
;        Author    : Mats Dal
;        Version   : 1.01
;        Date      : (YYMMDD)
;                   1.00 960724 : First release (DOS4G not working).
;                   1.01 960731 : DOS4G not a problem any more.
;
;        Descr    : RS232 interrupt handlers.
;;
              .486P
							.MODEL	FLAT
              
              .DATA

              EXTRN   C bitxfifo:     BYTE    ; How large is extern TX fifo.
              EXTRN   C TXtail:       WORD
              EXTRN   C TXhead:       WORD
              EXTRN   C RXhead:       WORD
              EXTRN   C TXfifo:       DATAPTR
              EXTRN   C RXfifo:       DATAPTR
              EXTRN   C RXfifoSize:   WORD
              EXTRN   C TXfifoSize:   WORD
              EXTRN   C RXfilled:     WORD
              EXTRN   C TXon:         BYTE
              EXTRN   C RTSon:        BYTE
              EXTRN   C TXdisabled:   BYTE
              EXTRN   C Msk8259:      BYTE
              EXTRN   C UARTbase:     WORD

IntrptAction  DD  int_modem, int_tx, int_rx, int_status

              .CODE

              EXTRN   __GETDS:PROC         ; Watcom internal procedure.

              PUBLIC  RS232_1hand_
              PUBLIC  RS232_2hand_
              PUBLIC  RS232_3hand_
              PUBLIC  RS232_4hand_
              
; Modemcontrol
DTR           EQU   1
RTS           EQU   2
OUT2          EQU   8

; ModemStatus
CTS           EQU   16

; Register offset UARTbase
IER           EQU   1
MCR           EQU   4

;;
;   Function: RS232_handler()
;   Type    : Interrupt
;   Descr   : Main interrupthandler.
;   Input   : None.
;   Output  : None.
;   Destroys: None.
;;
RS232_handler PROC  NEAR

    Push  ds
    Push  es
    Call  __GETDS                 ; Must set datasegment first.
    Mov   ax, ds
    Mov   es, ax

    In    al, 21h
    Mov   cl, Msk8259[ebx]
    Not   cl                      ; Set mask to off.
    Or    al, cl
    Out   21h, al                 ; Turn off comport nr. ebx UART interrupts.

    Mov   al, 20h                 ; Acknowledge interrupt.
    Out   20h, al

    Sti                           ; Permit other interrupts.

continue:
    Mov   dx, UARTbase[ebx*2]
    Add   dx, 2
    In    al, dx                  ; Get interrupt cause.

    Test  eax, 1                  ; Did the UART generate the interrupt?
    Jnz   SkipLoop

    And   eax, 6                  ; Mask bits not needed.
    Call  IntrptAction[eax*2]     ; Serve this interrupt. ebx == PORT.
    Jmp   continue                ; Look for more things to do.

SkipLoop:
  
    Cli                           ; Prohibit other to interrupt.

    In    al, 21h
    And   al, Msk8259[ebx]
    Out   21h, al                 ; Turn on UART 8259 interrupts.

    Pop   es
    Pop   ds
    Ret

RS232_handler ENDP

RS232_1hand_  PROC
    Pushad
    Mov   ebx, 0                  ; ebx = ComPort 1.
    Call  RS232_handler           ; Call main interrupt handler.
    Popad
    IRetd
RS232_1hand_  ENDP

RS232_2hand_  PROC
    Pushad
    Mov   ebx, 1                  ; ebx = ComPort 2.
    Call  RS232_handler           ; Call main interrupt handler.
    Popad
    IRetd
RS232_2hand_  ENDP

RS232_3hand_  PROC
    Pushad
    Mov   ebx, 2                  ; ebx = ComPort 3.
    Call  RS232_handler           ; Call main interrupt handler.
    Popad
    IRetd
RS232_3hand_  ENDP

RS232_4hand_  PROC
    Pushad
    Mov   ebx, 3                  ; ebx = ComPort 4.
    Call  RS232_handler           ; Call main interrupt handler.
    Popad
    IRetd
RS232_4hand_  ENDP

;;
;   Function: int_modem()
;   Type    : Procedure NEAR
;   Descr   : Gets modem status.
;   Input   : None.
;   Output  : None.
;   Destroys: DX, AL
;;
int_modem     PROC  NEAR

    Mov   dx, UARTbase[ebx*2]     ; Just clear.
    Add   dx, 6
    In    al, dx

    Mov   cl, al

    Test  cl, 00000001b           ; CTS has changed?
    Jz    NoCTSChange

    Mov   dx, UARTbase[ebx*2]
    Add   dx, IER                 ; dx = UARTbase[ebx*2] + IER
    Mov   al, 00001101b           ; al = disable transmit interrupts.
    Mov   TXon[ebx], 0            ; Show this to user also.

    Test  cl, 00010000b           ; CTS bit 4.
    Jz    StopTHRInter
    Mov   al, 00001111b           ; al = enable transmit interrupts.
    Mov   TXon[ebx], 1            ; Show this to user also.
StopTHRInter:
    Out   dx, al                  ; Do change.

NoCTSChange:
    Ret

int_modem     ENDP

;;
;   Function: int_tx()
;   Type    : Procedure NEAR
;   Descr   : Transmits possible chars in FIFO.
;   Input   : None.
;   Output  : None.
;   Destroys: None.
;;
int_tx        PROC  NEAR

    Mov   dx, UARTbase[ebx*2]
    Xor   esi, esi                ; Reset all bits in esi.
    Mov   si, TXtail[ebx*2]       ; esi = TXtail
    Mov   edi, TXfifo[ebx*4]      ; edi = TXfifo.

    Cmp   TXon[ebx], 0
    Je    CloseTX                 ; Skip transmit if TXon==0.

    Mov   cl, bitxfifo[ebx]
TXnext:
    Cmp   si, TXhead[ebx*2]
    Je    CloseTX                 ; more bytes to send?

    Mov   al, [edi+esi]           ; edi = TXfifo[port]
    Out   dx, al                  ; Write char.

    Inc   esi                     ; Update pointer.
    Cmp   si, TXfifoSize[ebx*2]   ; Wrapcheck.
    Jne   NoTXWrap
    Xor   esi, esi                ; Do wrap esi to zero.
NoTXWrap:
    Dec   cl
    Jnz   TXnext

    Jmp   EndTX

CloseTX:

    Add   dx, 1                   ; dx = UARTbase[ebx*2]
    Mov   al, 00001101b           ; Disable transmit interrupts.
    Out   dx, al
    Mov   TXdisabled[ebx], 1      ; Show this to user also.

EndTX:
    Mov   TXtail[ebx*2], si
    Ret

int_tx        ENDP

;;
;   Function: int_rx()
;   Type    : Procedure NEAR
;   Descr   : Recieve chars from UART and puts them in the FIFO.
;   Input   : None.
;   Output  : None.
;   Destroys: None.
;;
int_rx        PROC  NEAR

    Xor   esi, esi
    Mov   si, RXhead[ebx*2]
    Mov   cx, RXfilled[ebx*2]     ; Store RXfilled temporary in BX.
    Mov   edi, RXfifo[ebx*4]      ; edi = RXfifo[port].

RXnext:
    Cmp   cx, 32                  ; Must be larger than FIFOSIZE, 16.
    Jg    RXNotFull
; Warn, set RTS to low.
    Mov   dx, UARTbase[ebx*2]
    Add   dx, MCR                 ; dx = UARTbase[ebx*2] + MCR
    Mov   al, OUT2 + DTR          ; Set RTS to low.
    Out   dx, al
    Mov   RTSon[ebx], 0           ; Show this to user also.
RXNotFull:

    Mov   dx, UARTbase[ebx*2]
    In    al, dx                  ; Reads a character from UART.

    Test  cx, cx                  ; Check if RXfilled is zero,
    Jz    NoRXWrap                ; in other words if buffer is full!

    Dec   cx                      ; Update number of characters stored.
    Mov   [edi+esi], al           ; No, then store char in buffer.
    Inc   esi                     ; Update header.

    Cmp   si, RXfifoSize[ebx*2]   ; Check for wrap-around.
    Jne   NoRXWrap
    Xor   esi, esi                ; Yes, do wrap-around on header.
NoRXWrap:
    Mov   RXhead[ebx*2], si       ; Update transmit header.
    Mov   RXfilled[ebx*2], cx     ; Update RXfilled.

; Check if there is some more bytes to read.
    Mov   dx, UARTbase[ebx*2]
    Add   dx, 2
    In    al, dx                  ; Get interrupt cause.
    Test  eax, 1                  ; Did the UART generate the interrupt?
    Jnz   NoUartGen

    And   eax, 6                  ; Mask bits not needed.
    Cmp   eax, 4
    Je    RXnext                  ; Recieve some more. io FIFO not empty.
    Call  IntrptAction[eax*2]     ; Serve this interrupt.
NoUartGen:

    Mov   dx, UARTbase[ebx*2]
    Add   dx, 5
    In    al, dx
    Test  al, 40h                 ; Have we missed a THR interrupt?
    Jnz   int_tx                  ; Yes, serve that interrupt!
    Ret

int_rx        ENDP

;;
;   Function: int_status()
;   Type    : Procedure NEAR
;   Descr   : Reads line status change.
;   Input   : None.
;   Output  : None.
;   Destroys: None.
;;
int_status    PROC  NEAR

    Mov   dx, UARTbase[ebx*2]
    Add   dx, 5
    In    al, dx             ; Just read it to prevent locking.
    Ret

int_status    ENDP

              END

